package gov.usgs.earthquake.nshmp.calc;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.Text.LOG_INDENT;
import static gov.usgs.earthquake.nshmp.Text.NEWLINE;
import static gov.usgs.earthquake.nshmp.Text.NULL;
import static gov.usgs.earthquake.nshmp.geo.BorderType.MERCATOR_LINEAR;
import static gov.usgs.earthquake.nshmp.internal.Parsing.splitToList;
import static java.util.stream.Collectors.toUnmodifiableList;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Optional;

import com.google.common.collect.Iterables;

import gov.usgs.earthquake.nshmp.calc.Site.Builder;
import gov.usgs.earthquake.nshmp.geo.Bounds;
import gov.usgs.earthquake.nshmp.geo.GriddedRegion;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.Region;
import gov.usgs.earthquake.nshmp.geo.Regions;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.FeatureCollection;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson.Type;
import gov.usgs.earthquake.nshmp.geo.json.Properties;
import gov.usgs.earthquake.nshmp.internal.Parsing;
import gov.usgs.earthquake.nshmp.internal.Parsing.Delimiter;

/**
 * Iterable {@code Site} container. Factory methods support creating instances
 * from different input formats.
 *
 * @author U.S. Geological Survey
 */
public abstract class Sites implements Iterable<Site> {

  /**
   * Create an unmodifiable {@code Iterable<Site>} from the comma-delimted site
   * file designated by {@code path}.
   *
   * @param path to comma-delimited site data file
   * @throws IOException if a problem is encountered
   */
  public static Sites fromCsv(Path path, CalcConfig defaults) throws IOException {
    return fromCsv(path, defaults, false);
  }

  public static Sites fromCsv(Path path, CalcConfig defaults, boolean lenient) throws IOException {
    return readCsv(path, defaults, true);
  }

  private static Sites readCsv(Path path, CalcConfig defaults, boolean lenient) throws IOException {

    checkArgument(Files.exists(path), "Site file [%s] does not exist", path);

    List<Site> sites = new ArrayList<>();
    Builder siteBuilder = Site.builder(defaults);

    List<String> lines = Files.readAllLines(path, StandardCharsets.UTF_8);
    boolean firstline = true;
    List<String> keyList = new ArrayList<>();
    for (String line : lines) {

      /* Skip comments */
      if (line.startsWith("#")) {
        continue;
      }
      if (line.trim().isEmpty()) {
        continue;
      }

      List<String> values = Parsing.splitToList(line, Delimiter.COMMA);

      /* Set up key/column ordering */
      if (firstline) {
        for (String key : values) {
          if (!lenient) {
            checkState(Site.KEYS.contains(key), "Illegal site property key [%s]", key);
          }
          keyList.add(key);
        }
        checkState(keyList.contains(Site.Key.LAT), "Site latitudes must be defined");
        checkState(keyList.contains(Site.Key.LON), "Site longitudes must be defined");
        firstline = false;
        continue;
      }

      int index = 0;
      double lat = 0.0;
      double lon = 0.0;
      for (String key : keyList) {
        String value = values.get(index);
        switch (key) {
          case Site.Key.LAT:
            lat = Double.parseDouble(value);
            break;
          case Site.Key.LON:
            lon = Double.parseDouble(value);
            break;
          case Site.Key.NAME:
            siteBuilder.name(value);
            break;
          case Site.Key.VS30:
            siteBuilder.vs30(Double.parseDouble(value));
            break;
          case Site.Key.VS_INF:
            siteBuilder.vsInferred(Boolean.parseBoolean(value));
            break;
          case Site.Key.Z1P0:
            siteBuilder.z1p0(value.equals(NULL) ? Double.NaN : Double.parseDouble(value));
            break;
          case Site.Key.Z2P5:
            siteBuilder.z2p5(value.equals(NULL) ? Double.NaN : Double.parseDouble(value));
            break;
          default:
            if (!lenient) {
              throw new IllegalStateException("Unsupported site key: " + key);
            }
        }
        index++;
      }
      siteBuilder.location(lon, lat);
      sites.add(siteBuilder.build());
    }
    return new ListIterable(List.copyOf(sites));
  }

  /**
   * Create an unmodifiable {@code Iterable<Site>} from the GeoJSON site file
   * designated by {@code path}.
   *
   * @param path to GeoJson site data file
   * @throws IOException if a problem is encountered
   */
  public static Sites fromJson(Path path, CalcConfig defaults) throws IOException {
    FeatureCollection fc = GeoJson.from(path).toFeatureCollection();
    List<Feature> features = fc.features();
    if (features.get(0).type() == Type.POINT) {
      return createSiteList(features, defaults);
    } else {
      return createSiteRegion(features, defaults);
    }
  }

  private static Sites createSiteList(List<Feature> features, CalcConfig defaults) {
    List<Site> sites = features.stream()
        .map(feature -> Site.fromGeoJson(feature, defaults))
        .collect(toUnmodifiableList());
    return new ListIterable(sites);
  }

  /**
   * Create an unmodifiable singleton {@code Iterable<Site>} from the supplied
   * string. String is expected to have the form:
   *
   * <p>{@code "name,lon,lat[,vs30,vsInf[,z1p0,z2p5]]"}
   *
   * <p>Although the latter 4 fields are optional, if {@code vs30} is defined,
   * so too must {@code vsInf}. Likewise, if {@code z1p0} is defined, so too
   * must {@code z2p5}.
   *
   * @param s String to parse
   */
  public static Sites fromString(String s, CalcConfig defaults) {
    List<String> values = splitToList(s, Delimiter.COMMA);
    Builder b = Site.builder(defaults);
    checkArgument(
        values.size() > 2,
        NEWLINE + "    Site string \"%s\" is incomplete; it must contain at least 'name, lon, lat'",
        s);
    b.name(values.get(0)).location(
        Double.valueOf(values.get(1)),
        Double.valueOf(values.get(2)));
    if (values.size() > 3) {
      checkArgument(
          values.size() > 4,
          NEWLINE + "    Site string \"%s\" defines 'vs30'; it must also define 'vsInf'",
          s);
      b.vs30(Double.valueOf(values.get(3)));
      b.vsInferred(Boolean.valueOf(values.get(4)));
    }
    if (values.size() > 5) {
      checkArgument(
          values.size() > 6,
          NEWLINE + "    Site string \"%s\" defines 'z1p0'; it must also define 'z2p5'",
          s);
      String z1p0str = values.get(5);
      b.z1p0(z1p0str.equals(NULL) ? Double.NaN : Double.valueOf(z1p0str));
      String z2p5str = values.get(6);
      b.z2p5(z2p5str.equals(NULL) ? Double.NaN : Double.valueOf(z2p5str));
    }
    return new ListIterable(List.of(b.build()));
  }

  private static final int TO_STRING_LIMIT = 5;
  private static final String SITE_INDENT = LOG_INDENT + "       ";

  @Override
  public String toString() {
    boolean map = this instanceof RegionIterable;
    StringBuilder sb = new StringBuilder();
    if (map) {
      RegionIterable mapIterable = (RegionIterable) this;
      sb.append(mapIterable.region.name())
          .append(" Region [size=")
          .append(size())
          .append(" ").append(mapIterable.siteBuilder.state())
          .append("]");
    } else {
      sb.append("List")
          .append(" [size=")
          .append(size())
          .append("]");

      for (Site site : Iterables.limit(this, TO_STRING_LIMIT)) {
        sb.append(SITE_INDENT).append(site);
      }
      if (size() > TO_STRING_LIMIT) {
        int delta = size() - TO_STRING_LIMIT;
        sb.append(SITE_INDENT).append("... and ").append(delta).append(" more ...");
      }
    }
    return sb.toString();
  }

  /**
   * The number of {@code Site}s {@code this} contains.
   */
  public abstract int size();

  /**
   * An optional {@code Bounds} that is used to specify rectangular map extents,
   * which may differ from the range spanned by the sites in this. Presently
   * only containers initialized with GeoJSON that includes an 'Extents' polygon
   * will return an {@code Optional} where {@link Optional#isPresent()}
   * {@code == true}.
   */
  public abstract Optional<Bounds> mapBounds();

  /**
   * An optional spacing value that reflects the spacing of sites within an
   * attendant {@link #mapBounds()}.
   */
  public abstract Optional<Double> mapSpacing();

  /*
   * TODO consider: if csv or geojson files are long, and if basinDataProvider
   * is being used, all calls to data provider must be made to create list to
   * init Iterable; however this would probably necessitate passing in a list of
   * dummy sites, the values of which could be overridden by the data provider
   */

  private static final class ListIterable extends Sites {
    final List<Site> delegate;

    ListIterable(List<Site> delegate) {
      this.delegate = delegate;
    }

    @Override
    public Iterator<Site> iterator() {
      return delegate.iterator();
    }

    @Override
    public int size() {
      return delegate.size();
    }

    @Override
    public Optional<Bounds> mapBounds() {
      return Optional.empty();
    }

    @Override
    public Optional<Double> mapSpacing() {
      return Optional.empty();
    }
  }

  private static final class RegionIterable extends Sites {

    final GriddedRegion region;
    final Builder siteBuilder;
    final Optional<Bounds> bounds;

    RegionIterable(
        GriddedRegion region,
        Builder siteBuilder,
        Optional<Bounds> bounds) {

      this.region = region;
      this.siteBuilder = siteBuilder;
      this.bounds = bounds;
    }

    @Override
    public Iterator<Site> iterator() {
      return new Iterator<Site>() {
        final Iterator<Location> locations = region.iterator();

        @Override
        public boolean hasNext() {
          return locations.hasNext();
        }

        @Override
        public Site next() {
          siteBuilder.location(locations.next());
          return siteBuilder.build();
        }

        @Override
        public void remove() {
          throw new UnsupportedOperationException();
        }
      };
    }

    @Override
    public int size() {
      return region.size();
    }

    @Override
    public Optional<Bounds> mapBounds() {
      return bounds;
    }

    @Override
    public Optional<Double> mapSpacing() {
      return Optional.of(region.latSpacing());
    }
  }

  private static Sites createSiteRegion(List<Feature> features, CalcConfig defaults) {
    checkState(features.size() <= 2, "Only 2 polygon features may be defined");

    Optional<Bounds> mapBounds = Optional.empty();
    String boundsName = "";

    /* Possibly set map extents. */
    int mapRegionIndex = 0;
    if (features.size() > 1) {
      mapRegionIndex = 1;
      Feature mapPoly = features.get(0);
      LocationList mapPolyBorder = mapPoly.asPolygonBorder();
      mapBounds = Optional.of(validateExtents(mapPolyBorder).bounds());
      boundsName = mapPoly.properties().getString(Site.Key.NAME).orElse("Map Extents");
    }

    Feature sitesPoly = features.get(mapRegionIndex);
    LocationList sitesPolyBorder = sitesPoly.asPolygonBorder();
    Properties properties = sitesPoly.properties();
    String mapName = properties.getString(Site.Key.NAME).orElse("Unnamed Map");

    /*
     * We special case a 5-coordinate border that defines a mercator recangle so
     * as to create a region that includes sites on the north and east borders.
     */

    Region calcRegion = null;
    try {
      Bounds b = validateExtents(sitesPolyBorder).bounds();
      calcRegion = Regions.createRectangular(mapName, b.min, b.max);
    } catch (IllegalArgumentException iae) {
      calcRegion = Regions.create(mapName, sitesPolyBorder, MERCATOR_LINEAR);
    }

    double spacing = properties.getDouble("spacing").orElseThrow();

    /* Builder used to create all sites when iterating over region. */
    Builder builder = Site.builder(defaults)
        .geoJsonProperties(properties);

    Region mapRegion = calcRegion;
    if (mapBounds.isPresent()) {
      Bounds b = mapBounds.orElseThrow();
      Region r = Regions.createRectangular(boundsName, b.min, b.max);
      mapRegion = Regions.intersectionOf(mapName, r, calcRegion);
    }

    GriddedRegion region = Regions.toGridded(
        mapRegion,
        spacing, spacing,
        GriddedRegion.ANCHOR_0_0);

    return new RegionIterable(region, builder, mapBounds);
  }

  private static LocationList validateExtents(LocationList locations) {
    // TODO is this a linearRing check?? answer: sort of; need 3-pt check
    // TODO should/could add winding check
    // TODO also need check for

    /* Recangular linear ring check. */
    checkArgument(
        locations.size() == 5,
        "Extents polygon must contain 5 coordinates: %s",
        locations);

    /* Order agnostic rectilinear lat-lon check. */
    Location p1 = locations.get(0);
    Location p2 = locations.get(1);
    Location p3 = locations.get(2);
    Location p4 = locations.get(3);
    boolean rectangular = (p1.latitude == p2.latitude)
        ? (p3.latitude == p4.latitude &&
            p1.longitude == p4.longitude &&
            p2.longitude == p3.longitude)
        : (p1.latitude == p4.latitude &&
            p2.latitude == p3.latitude &&
            p1.longitude == p2.longitude &&
            p3.longitude == p4.longitude);
    checkArgument(
        rectangular,
        "Extents polygon does not define a lat-lon Mercator rectangle: %s",
        locations);

    return locations;
  }

}
