package gov.usgs.earthquake.nshmp.internal.www;

import java.lang.reflect.Type;
import java.time.format.DateTimeFormatter;
import java.util.Optional;

import com.google.common.collect.Range;
import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonPrimitive;
import com.google.gson.JsonSerializationContext;
import com.google.gson.JsonSerializer;

import gov.usgs.earthquake.nshmp.gmm.GmmInput;
import gov.usgs.earthquake.nshmp.gmm.GmmInput.Field;
import gov.usgs.earthquake.nshmp.internal.www.meta.ParamType;

/**
 * Web service utilities.
 *
 * @author U.S. Geological Survey
 */
public class WsUtils {

  public static final DateTimeFormatter DATE_FMT = DateTimeFormatter.ofPattern(
      "yyyy-MM-dd'T'HH:mm:ssXXX");

  public static <T, E extends Enum<E>> T checkValue(E key, T value) {
    if (value == null) {
      throw new IllegalStateException("Missing [" + key.toString() + "]");
    }

    return value;
  }

  /* Constrain all doubles to 8 decimal places */
  public static final class DoubleSerializer implements JsonSerializer<Double> {
    @Override
    public JsonElement serialize(Double d, Type type, JsonSerializationContext context) {
      double dOut = Double.valueOf(String.format("%.8g", d));
      return new JsonPrimitive(dOut);
    }
  }

  /* Serialize param type enum as lowercase */
  public static class ParamTypeSerializer implements JsonSerializer<ParamType> {
    @Override
    public JsonElement serialize(ParamType paramType, Type type, JsonSerializationContext context) {
      return new JsonPrimitive(paramType.name().toLowerCase());
    }
  }

  /* Convert NaN to null */
  public static final class NaNSerializer implements JsonSerializer<Double> {
    @Override
    public JsonElement serialize(Double d, Type type, JsonSerializationContext context) {
      return Double.isNaN(d) ? null : new JsonPrimitive(d);
    }
  }

  public static final class ConstraintsSerializer implements JsonSerializer<GmmInput.Constraints> {
    @Override
    public JsonElement serialize(
        GmmInput.Constraints constraints,
        Type type,
        JsonSerializationContext context) {
      JsonArray json = new JsonArray();

      for (Field field : Field.values()) {
        Optional<?> opt = constraints.get(field);
        if (opt.isPresent()) {
          Range<?> value = (Range<?>) opt.orElseThrow();
          Constraint constraint = new Constraint(
              field.id,
              value.lowerEndpoint(),
              value.upperEndpoint());
          json.add(context.serialize(constraint));
        }
      }

      return json;
    }
  }

  public static final class EnumSerializer<E extends Enum<E>> implements JsonSerializer<E> {

    @Override
    public JsonElement serialize(E src, Type type, JsonSerializationContext context) {

      int displayOrder = src.ordinal();

      JsonObject jObj = new JsonObject();
      jObj.addProperty("id", src.ordinal());
      jObj.addProperty("value", src.name());
      jObj.addProperty("display", src.toString());
      jObj.addProperty("displayorder", displayOrder);

      return jObj;
    }
  }

  @SuppressWarnings("unused")
  private static class Constraint {
    final String id;
    final Object min;
    final Object max;

    Constraint(String id, Object min, Object max) {
      this.id = id;
      this.min = min;
      this.max = max;
    }
  }

}
