package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static com.google.common.base.StandardSystemProperty.LINE_SEPARATOR;
import static gov.usgs.earthquake.nshmp.Text.checkName;
import static gov.usgs.earthquake.nshmp.model.SourceType.FAULT_CLUSTER;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.Locations;

/**
 * Cluster source representation. Each cluster source wraps a
 * {@code FaultSourceSet} containing one or more fault representations that
 * rupture as independent events but with a similar rate. For example, at New
 * Madrid, each ClusterSourceSet has 5 ClusterSources, one for each position
 * variant of the model. For each position variant there is one FaultSourceSet
 * containing the FaultSources in the cluster, each of which may have one, or
 * more, magnitude or other variants represented by its internal list of
 * {@code IncrementalMfd}s.
 *
 * <p>Cluster source hazard is calculated from the joint probabilities of ground
 * motions from the wrapped faults, which is handled internally by a separate
 * calculator and {@link ClusterRuptureSet#iterator()} therefore throws an
 * {@code UnsupportedOperationException}.
 *
 * <p>Unlike other {@code Source}s whose weights are carried exclusively with
 * their associated {@link IncrementalMfd}, {@code ClusterSource}s carry an
 * additional {@link #weight()} value.
 *
 * <p>A {@code ClusterSource} cannot be created directly; it may only be created
 * by a private parser.
 *
 * @author U.S. Geological Survey
 */
public class ClusterRuptureSet implements RuptureSet {

  final String name;
  final int id;

  final ModelData data; // holds rate tree (numeric values)
  final List<FaultRuptureSet> faultRuptureSets;

  private ClusterRuptureSet(Builder builder) {
    this.name = builder.name;
    this.id = builder.id;
    this.data = builder.data;
    this.faultRuptureSets = builder.faultRuptureSets;
  }

  @Override
  public String name() {
    return name;
  }

  @Override
  public int id() {
    return id;
  }

  @Override
  public int size() {
    return faultRuptureSets.size();
  }

  @Override
  public SourceType type() {
    return FAULT_CLUSTER;
  }

  /**
   * The closest point across the traces of all fault sources that participate
   * in this cluster, relative to the supplied site {@code Location}.
   */
  @Override
  public Location location(Location site) {
    LocationList.Builder locs = LocationList.builder();
    for (FaultRuptureSet fault : faultRuptureSets) {
      locs.add(fault.location(site));
    }
    return Locations.closestPoint(site, locs.build());
  }

  /**
   * The weight applicable to this {@code ClusterSource}.
   */
  public double weight() {
    return 1.0; // TODO wrong
  }

  /**
   * The {@code FaultSourceSet} of all {@code FaultSource}s that participate in
   * this cluster.
   */
  public FaultSourceSet faults() {
    return null; // TODO needed by rate calculation
  }

  @Override
  public String toString() {
    Map<String, ?> data = Map.of(
        "name", name(),
        "weight", weight());
    StringBuilder sb = new StringBuilder()
        .append(getClass().getSimpleName())
        .append(" ")
        .append(data)
        .append(LINE_SEPARATOR.value());
    for (FaultRuptureSet frs : faultRuptureSets) {
      sb.append("  ")
          .append(frs.toString())
          .append(LINE_SEPARATOR.value());
    }
    return sb.toString();
  }

  static Builder builder() {
    return new Builder();
  }

  /* Single use builder */
  static class Builder {

    boolean built = false;

    private ModelData data; // holds cluster rate-tree
    private String name;
    private Integer id;

    List<FaultRuptureSet> faultRuptureSets = new ArrayList<>();

    Builder name(String name) {
      this.name = checkName(name, "ClusterRuptureSet");
      return this;
    }

    Builder id(int id) {
      this.id = id;
      return this;
    }

    /* Set MFD helper data. */
    Builder modelData(ModelData data) {
      this.data = data;
      return this;
    }

    Builder addRuptureSet(FaultRuptureSet ruptureSet) {
      if (faultRuptureSets.size() > 1) {
        MfdTrees.checkTreeIdsAndWeights(
            faultRuptureSets.get(0).mfdTree,
            ruptureSet.mfdTree);
      }
      faultRuptureSets.add(ruptureSet);
      return this;
    }

    void validateState(String label) {
      checkState(!built, "Single use builder");
      checkNotNull(name, "%s name", label);
      checkNotNull(id, "%s id", label);
      checkNotNull(data, "%s model data", label);
      checkState(faultRuptureSets.size() > 0);
      built = true;
    }

    ClusterRuptureSet build() {
      validateState("ClusterSource.Builder");
      return new ClusterRuptureSet(this);
    }
  }

}
