package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.Earthquakes.MAG_RANGE;
import static gov.usgs.earthquake.nshmp.model.SourceType.FAULT;
import static java.util.stream.Collectors.toUnmodifiableList;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import com.google.common.collect.Iterables;

import gov.usgs.earthquake.nshmp.data.XyPoint;
import gov.usgs.earthquake.nshmp.fault.surface.DefaultGriddedSurface;
import gov.usgs.earthquake.nshmp.fault.surface.GriddedSurface;
import gov.usgs.earthquake.nshmp.fault.surface.RuptureFloating;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Fault source representation. This class wraps a model of a fault geometry and
 * a list of magnitude frequency distributions that characterize how the fault
 * might rupture (e.g. as one, single geometry-filling event, or as multiple
 * smaller events) during earthquakes. Smaller events are modeled as 'floating'
 * ruptures; they occur in multiple locations on the fault surface with
 * appropriately scaled rates.
 *
 * <p>A {@code FaultSource} cannot be created directly; it may only be created
 * by a private parser.
 *
 * @author U.S. Geological Survey
 */
public class FaultSource implements Source {

  final SourceFeature.NshmFault feature;
  final SourceConfig.Fault config;
  final GriddedSurface surface;

  final LogicTree<Mfd> mfdTree;
  final double branchWeight; // rupture set branch weight

  private final List<Mfd> scaledMfds; // includes ruptureSet weight
  private final List<List<Rupture>> ruptureLists; // 1:1 with Mfds

  FaultSource(Builder builder) {
    this.feature = builder.feature;
    this.config = builder.config;
    this.mfdTree = builder.mfdTree;
    this.branchWeight = builder.branchWeight;

    this.surface = DefaultGriddedSurface.builder()
        .trace(feature.trace)
        .depth(feature.upperDepth)
        .dip(feature.dip)
        .width(feature.width)
        .spacing(config.surfaceSpacing)
        .build();

    this.scaledMfds = MfdTrees.scaledMfdList(mfdTree, branchWeight);
    // System.out.println(mfdTree);
    this.ruptureLists = scaledMfds.stream()
        .map(this::createRuptureList)
        .collect(toUnmodifiableList());

    // scaledMfds.stream()
    // .forEach(System.out::println);

    boolean noRuptures = Iterables.size(Iterables.concat(ruptureLists)) == 0;
    if (noRuptures) {
      System.out.println("      No ruptures: " + feature.name);
    }

    // TODO refactor to use size()
    // checkState(Iterables.size(Iterables.concat(ruptureLists)) > 0,
    // "FaultSource has no ruptures");

    // for (Mfd mfd : scaledMfds) {
    // System.out.println(mfd.data());
    // }
    // System.out.println(scaledMfds);
    // System.out.println(ruptureLists.size());

    // TODO clean
    // scaledMfds.stream().map(Mfd::data).forEach(System.out::println);

    // TODO fault source total MFD not yet scaled by source weight (e.g. dip
    // branch) TODO at least for 2014 the MFD tree and total MFD are identical
    // across dip branches

    // System.out.println(this);

    // System.out.println(scaledMfds);
    // long nRuptures = ruptureLists.stream().flatMap(List::stream).count();
    // System.out.println(nRuptures);

    // System.out.println(Mfds.momentRate(scaledMfds.get(0)));
    // System.out.println(trace);
    // System.out.println(surface);
  }

  @Override
  public String name() {
    return feature.name;
  }

  @Override
  public int id() {
    return feature.id;
  }

  @Override
  public SourceType type() {
    return FAULT;
  }

  @Override
  public int size() {
    return ruptureLists.stream()
        .mapToInt(List::size)
        .sum();
  }

  /**
   * The closest point on the fault trace, relative to the supplied site
   * {@code Location}.
   */
  @Override
  public Location location(Location site) {
    return Locations.closestPoint(site, feature.trace);
  }

  @Override
  public List<Mfd> mfds() {
    return mfdTree.stream()
        .map(Branch::value)
        .collect(toUnmodifiableList());
  }

  @Override
  public Iterator<Rupture> iterator() {
    return Iterables.concat(ruptureLists).iterator();
  }

  @Override
  public String toString() {
    Map<String, ?> data = Map.of(
        "name", feature.name,
        "dip", feature.dip,
        "width", feature.width,
        "rake", feature.rake,
        "mfds", mfdTree.size(),
        "top", feature.trace.first().depth);
    return getClass().getSimpleName() + " " + data;
  }

  private List<Rupture> createRuptureList(Mfd mfd) {

    List<Rupture> ruptureList = new ArrayList<>();

    /*
     * Responsibility for setting floatRuptures is passed to builders; interface
     * and fault sources will have separate rules governing when to float.
     *
     * If floatRuptures is false, rupture list builder will just create a single
     * rupture for each magnitude in the supplied MFD.
     *
     * If floatRuptures is true, rupture list builder will make an additional
     * check that rupture length is shorter than the trace and only float those
     * that are shorter.
     *
     * For the above rules to work, we need to be careful about combining mixed
     * MFD types. For example, in the WUS we'll want to create independent
     * rupture sets for SINGLE (floats = false, usually) and GR (floats = true,
     * always) ruptures. Under some circumstances, SINGLE ruptures will float
     * (e.g., regional, a-priori, or historic-eq-based mMax < mMax from scaling
     * relation)
     *
     * 2014/18 Search model for floats="true"; fixed M 7.0 and 7.5 usually
     * float; 'partial ruptures' (GR) for which M<6.55 are modeled as floating
     * SINGLE
     *
     * need to add regional mMax
     *
     * move floats flag to configs?
     */

    /*
     * Width is irrelevant as we know current WUS models only use WC-Length.
     *
     * TODO need to check that 'characteristic' magnitudes computed to set
     * MfdProperties are then (round-trip) consistent with lengths computed
     * here; considering the same trace is used values should be exactly equal.
     *
     * CONSIDER JUST INCLUDING THE FLOATS PROPERTY IN MfdProperties.SINGLE, why
     * not? it would be Optional.empty() for ruptures with prescriptive single
     * MFD's new Madrid and only get populated when deriving MFDs from slip
     * rates in the WUS. What happens in this situation if we've a logic tree
     * magnitude scaling relations.
     *
     * 7/30/2020 can specify based on source type whether to float single
     * ruptures; ruptures will always float if GR
     */

    // double mMaxGlobal = Earthquakes.MAG_RANGE.upperEndpoint();

    double magRef = feature.magnitude.orElse(MAG_RANGE.upperEndpoint());

    double refLength = config.ruptureScaling.dimensions(magRef, feature.width).length;
    double traceLength = feature.trace.length();
    // double mRef = feature.magnitude.orElse(Double.MAX_VALUE);

    // System.out.println(" mRef: " + magRef);
    // System.out.println(" refLength: " + refLength);
    // System.out.println("traceLength: " + traceLength);

    Mfd.Type mfdType = mfd.properties().type();

    // if (feature.magnitude.isPresent()) {
    // double mLength = config.ruptureScaling.dimensions(feature.magnitude,
    // feature.width).length;
    // }
    // TODO if the original SINGLE magnitude w/o any uncertainty is shorter than
    // than the trace length, then we float

    // System.out.println(feature.magnitude.);

    // System.out.println("Mw: " + magRef);

    /*
     * First float check; use reference length from source to determine if true
     * for SINGLE MFDs, otherwise epistemic branches end up floating.
     */
    boolean otherNoFloat = feature.magnitudeNote.isPresent() &&
        feature.magnitudeNote.orElseThrow().equals(MagnitudeNote.OTHER_NO_FLOAT);
    boolean floats = (mfdType == Mfd.Type.GR) ||
        (mfdType == Mfd.Type.SINGLE && refLength < traceLength && !otherNoFloat);

    for (XyPoint xy : mfd.data()) {
      double mag = xy.x();
      double rate = xy.y();

      // }
      // for (int i = 0; i < mfd.getNum(); ++i) {
      // double mag = mfd.getX(i);
      // double rate = mfd.getY(i);

      // TODO do we really want to do this??
      // TODO low rate shortcut should be derived from config
      // and applied when building input lists
      if (rate < 1e-14) {
        continue; // shortcut low rates
      }

      // TODO should width here pull directly from feature.width
      double mMfdLength = config.ruptureScaling.dimensions(mag, surface.width()).length;

      // TODO we want to get the 'floats' attribute out of MFDs
      // the only reason it is there is to allow SINGLE to flip-flop
      // it should just be a SourceProperty

      RuptureFloating floating = (floats && mMfdLength < traceLength)
          ? config.ruptureFloating
          : RuptureFloating.OFF;

      List<Rupture> floaters = floating.createFloatingRuptures(
          surface,
          config.ruptureScaling,
          mag,
          rate,
          feature.rake);

      ruptureList.addAll(floaters);

      // if (floats && msrLength < traceLength) {
      //
      // // TODO change to using RuptureFloating.OFF
      //
      // List<Rupture> floaters = config.ruptureFloating.createFloatingRuptures(
      // surface,
      // config.ruptureScaling,
      // mag,
      // rate,
      // feature.rake);
      // ruptureList.addAll(floaters);
      //
      // } else {
      // Rupture rup = Rupture.create(mag, rate, feature.rake, surface);
      // ruptureList.add(rup);
      // }
    }

    // We now have rates of zero for null branches that preserve
    // logic tree structure but that result in empty rupture lists
    // checkState(ruptureList.size() > 0, "Rupture list for %s is empty", name);

    return ruptureList; // ends up in collected unmodifiable list
  }

  /* Single use builder */
  static class Builder {

    private static final String ID = "FaultSource.Builder";
    private boolean built = false;

    SourceFeature.NshmFault feature;
    SourceConfig.Fault config;
    LogicTree<Mfd> mfdTree;
    Mfd mfdTotal;

    Boolean floatSingleRuptures;
    Double branchWeight;

    Builder feature(SourceFeature.NshmFault feature) {
      this.feature = feature;
      return this;
    }

    Builder config(SourceConfig.Fault config) {
      this.config = config;
      return this;
    }

    Builder mfdTree(LogicTree<Mfd> mfdTree) {
      checkArgument(checkNotNull(mfdTree).size() > 0, "MFD list is empty");
      this.mfdTree = mfdTree;
      return this;
    }

    Builder mfdTotal(Mfd mfdTotal) {
      checkNotNull(mfdTotal, "MFD total is null");
      this.mfdTotal = mfdTotal;
      return this;
    }

    Builder branchWeight(double weight) {
      this.branchWeight = weight;
      return this;
    }

    FaultSource build() {
      checkState(!built, "This %s instance as already been used", ID);
      checkState(feature != null, "%s feature not set", ID);
      checkState(config != null, "%s config not set", ID);
      checkState(mfdTree != null, "%s MFD tree not set", ID);
      checkState(mfdTotal != null, "%s total MFD not set", ID);
      checkState(branchWeight != null, "%s source branch weight not set", ID);
      built = true;
      return new FaultSource(this);
    }
  }
}
