package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.model.Deserialize.MFD_TREE;
import static gov.usgs.earthquake.nshmp.model.SourceType.INTERFACE;
import static java.util.stream.Collectors.collectingAndThen;
import static java.util.stream.Collectors.toList;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import gov.usgs.earthquake.nshmp.fault.surface.ApproxGriddedSurface;
import gov.usgs.earthquake.nshmp.fault.surface.GriddedSurface;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Subduction interface rupture set.
 *
 * @author U.S. Geological Survey
 */
public class InterfaceRuptureSet implements RuptureSet {

  /* May be original interface section or stitched. */
  final SourceFeature.Interface feature;
  final ModelData data;

  final LogicTree<Mfd> mfdTree;
  final Mfd mfdTotal;

  final List<Integer> sectionIds;// reference: actually needed?

  final GriddedSurface surface;

  InterfaceRuptureSet(Builder builder) {
    this.feature = builder.feature;
    this.data = builder.data;

    this.mfdTree = builder.mfdTree;
    this.mfdTotal = builder.mfdTotal;

    this.sectionIds = builder.sectionIds;
    this.surface = builder.surface;
  }

  @Override
  public String name() {
    return feature.name;
  }

  @Override
  public int id() {
    return feature.id;
  }

  @Override
  public int size() {
    // TODO what should this be ??
    // a value based on the size of the Mfds??
    return 1;
  }

  @Override
  public SourceType type() {
    return INTERFACE;
  }

  /**
   * The closest point on the fault trace, relative to the supplied site
   * {@code Location}.
   */
  @Override
  public Location location(Location site) {
    // TODO update to all traces
    return Locations.closestPoint(site,
        LocationList.builder()
            .addAll(feature.traces.get(0))
            .addAll(feature.traces.get(1))
            .build());
  }

  @Override
  public String toString() {
    Map<Object, Object> data = Map.of(
        "name", name(),
        "id", id(),
        "feature", feature.source.toJson());
    return getClass().getSimpleName() + " " + data;
  }

  static Builder builder() {
    return new Builder();
  }

  /* Single use builder */
  static class Builder {

    private boolean built = false;

    private SourceFeature.Interface feature;
    private ModelData data;
    private List<Integer> sectionIds;

    private LogicTree<Mfd.Properties> mfdPropsTree;

    /* created on build */
    private LogicTree<Mfd> mfdTree;
    private Mfd mfdTotal;
    private GriddedSurface surface;

    /* Validated and carried forward in rupture set feature. */
    private String name;
    private Integer id;

    Builder name(String name) {
      this.name = name;
      return this;
    }

    Builder id(int id) {
      this.id = id;
      return this;
    }

    Builder sections(List<Integer> sections) {
      checkArgument(sections.size() >= 1);
      this.sectionIds = List.copyOf(sections);
      return this;
    }

    /* Set the MFD tree. */
    Builder mfdTree(LogicTree<Mfd.Properties> mfdTree) {
      this.mfdPropsTree = mfdTree;
      return this;
    }

    /* Set MFD helper data. */
    Builder modelData(ModelData data) {
      this.data = data;
      return this;
    }

    InterfaceRuptureSet build() {
      validateAndInit("InterfaceRuptureSet.Builder");
      return new InterfaceRuptureSet(this);
    }

    private void validateAndInit(String label) {
      checkState(!built, "Single use builder");
      checkNotNull(name, "%s name", label);
      checkNotNull(id, "%s id", label);

      checkNotNull(data, "%s model data", label);
      checkNotNull(sectionIds, "%s section id list", label);
      checkNotNull(mfdPropsTree, "%s MFD logic tree", label);

      // TODO temp checking props are always the same;
      // don't think this is always true at least for faults
      checkState(mfdPropsTree.stream()
          .map(Branch::value)
          .map(Mfd.Properties::type)
          .distinct().limit(2).count() <= 1);

      mfdTree = buildMfdTree(data, mfdPropsTree);
      mfdTotal = MfdTrees.reduceMfdTree(mfdTree);

      feature = createFeature();

      SourceConfig.Interface config = data.sourceConfig().asInterface();
      surface = new ApproxGriddedSurface(
          feature.traces.get(0),
          feature.traces.get(1),
          config.surfaceSpacing);

      built = true;
    }

    private SourceFeature.Interface createFeature() {
      Map<Integer, SourceFeature.Interface> featureMap =
          data.interfaceFeatureMap().orElseThrow();
      // System.out.println(sectionIds);
      // System.out.println(featureMap.keySet());
      checkState(featureMap.keySet().containsAll(sectionIds));

      Feature f = joinFeatures(sectionIds.stream()
          .map(featureMap::get)
          .map(feature -> feature.source)
          .collect(toList()));
      return new SourceFeature.Interface(f, false);
    }

    /*
     * Join the traces in the supplied features into a new Feature and set name
     * and id of rupture-set. At present, interface sections have no geologic
     * properties such that we need to pick up properties from the first section
     * or override them with properties specified for the rupture-set. Moreover,
     * all current NSHM interface rupture-set sections have matching endpoints
     * and do not require coordinate/trace overrides.
     */
    private Feature joinFeatures(List<Feature> features) {
      Feature.Builder feature = (features.size() == 1)
          ? Feature.copyOf(features.get(0))
          : joinInterfaceSections(features);
      feature.id(id).properties(Map.of("name", name));
      return feature.build();
    }

    /*
     * Join the traces of the supplied sections in order, removing duplicate
     * points at the ends of sections.
     */
    private static Feature.Builder joinInterfaceSections(List<Feature> features) {
      int traceCount = features.get(0).asMultiLineString().size();
      List<LocationList> joinedTraces = new ArrayList<>();
      for (int i = 0; i < traceCount; i++) {
        LocationList trace = joinTracesAtIndex(features, i);
        joinedTraces.add(trace);
      }
      return Feature.multiLineString(joinedTraces);
    }

    /*
     * Join traces nested in MultiLineString geometries, throwing
     * IndexOutOfBoundsException if trace counts are not the same.
     */
    private static LocationList joinTracesAtIndex(List<Feature> features, int index) {
      return features.stream()
          .map(Feature::asMultiLineString)
          .map(list -> list.get(index))
          .flatMap(LocationList::stream)
          .distinct()
          .collect(collectingAndThen(
              toList(),
              LocationList::copyOf));
    }
  }

  private static LogicTree<Mfd> buildMfdTree(
      ModelData mfdData,
      LogicTree<Mfd.Properties> mfdPropertiesTree) {

    LogicTree.Builder<Mfd> mfdTree = LogicTree.builder(MFD_TREE);

    for (Branch<Mfd.Properties> branch : mfdPropertiesTree) {
      switch (branch.value().type()) {
        case SINGLE:
          createSingleMfds(mfdTree, mfdData, branch);
          break;
        case GR:
          createGrMfds(mfdTree, mfdData, branch);
          break;
        default:
          throw new UnsupportedOperationException(
              branch.value().type() + " MFD not supported");
      }
    }
    return mfdTree.build();
  }

  /* Interface GR */
  private static void createGrMfds(
      LogicTree.Builder<Mfd> mfdTree,
      ModelData mfdData,
      Branch<Mfd.Properties> branch) {

    // TODO add Mfd.Properties.toMfd() for when we know we have an a-value
    Mfd.Properties.GutenbergRichter props = branch.value().getAsGr();
    Mfd mfd = props.toBuilder().build();
    mfdTree.addBranch(branch.id(), mfd, branch.weight());
  }

  /* Interface SINGLE */
  private static void createSingleMfds(
      LogicTree.Builder<Mfd> mfdTree,
      ModelData mfdData,
      Branch<Mfd.Properties> branch) {

    Mfd.Properties.Single props = branch.value().getAsSingle();
    /* rates deserilized from json; may not always hold */
    Mfd mfd = props.toBuilder().build();
    mfdTree.addBranch(branch.id(), mfd, branch.weight());
  }

}
