package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;

import java.nio.file.Path;
import java.util.Map;
import java.util.Optional;

import gov.usgs.earthquake.nshmp.fault.FocalMech;
import gov.usgs.earthquake.nshmp.fault.surface.RuptureFloating;
import gov.usgs.earthquake.nshmp.fault.surface.RuptureScaling;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/*
 * Source model configuration wrapper class.
 *
 * @author U.S. Geological Survey
 */
abstract class SourceConfig {

  final Path resource;

  private SourceConfig(Path resource) {
    this.resource = resource;
  }

  Fault asFault() {
    return (Fault) this;
  }

  Grid asGrid() {
    return (Grid) this;
  }

  Interface asInterface() {
    return (Interface) this;
  }

  static class Fault extends SourceConfig {

    final double surfaceSpacing;
    final RuptureFloating ruptureFloating;
    final RuptureScaling ruptureScaling;
    final Optional<LogicTree<Double>> dipTree;
    final Optional<DipSlipModel> dipSlipModel;

    private Fault(Builder builder) {
      super(builder.resource);
      this.surfaceSpacing = builder.surfaceSpacing;
      this.ruptureFloating = builder.ruptureFloating;
      this.ruptureScaling = builder.ruptureScaling;
      this.dipTree = builder.dipTree;
      this.dipSlipModel = builder.dipSlipModel;
    }

    static Builder builder(Path resource) {
      return new Builder(resource);
    }

    static class Builder {

      private boolean built = false;

      private Path resource;
      private Double surfaceSpacing;
      private RuptureFloating ruptureFloating;
      private RuptureScaling ruptureScaling;
      private Optional<LogicTree<Double>> dipTree = Optional.empty();
      private Optional<DipSlipModel> dipSlipModel = Optional.empty();

      private Builder(Path resource) {
        this.resource = resource;
      }

      Builder spacing(double spacing) {
        checkArgument(spacing > 0.0);
        this.surfaceSpacing = spacing;
        return this;
      }

      Builder floating(RuptureFloating floating) {
        this.ruptureFloating = floating;
        return this;
      }

      Builder scaling(RuptureScaling scaling) {
        this.ruptureScaling = scaling;
        return this;
      }

      Builder dipTree(LogicTree<Double> dipTree) {
        this.dipTree = Optional.of(dipTree);
        return this;
      }

      Builder dipSlipModel(DipSlipModel dipSlipModel) {
        this.dipSlipModel = Optional.of(dipSlipModel);
        return this;
      }

      Fault build() {
        checkState(!built, "This builder has already been used");
        checkNotNull(resource);
        checkNotNull(surfaceSpacing);
        checkNotNull(ruptureFloating);
        checkNotNull(ruptureScaling);
        this.built = true;
        return new Fault(this);
      }
    }
  }

  static class Interface extends SourceConfig {

    final double surfaceSpacing;
    final RuptureFloating ruptureFloating;
    final RuptureScaling ruptureScaling;

    private Interface(Builder builder) {
      super(builder.resource);
      this.surfaceSpacing = builder.surfaceSpacing;
      this.ruptureFloating = builder.ruptureFloating;
      this.ruptureScaling = builder.ruptureScaling;
    }

    static Builder builder(Path resource) {
      return new Builder(resource);
    }

    static class Builder {

      private boolean built = false;

      private Path resource;
      private Double surfaceSpacing;
      private RuptureFloating ruptureFloating;
      private RuptureScaling ruptureScaling;

      private Builder(Path resource) {
        this.resource = resource;
      }

      Builder spacing(double spacing) {
        checkArgument(spacing > 0.0);
        this.surfaceSpacing = spacing;
        return this;
      }

      Builder floating(RuptureFloating floating) {
        this.ruptureFloating = floating;
        return this;
      }

      Builder scaling(RuptureScaling scaling) {
        this.ruptureScaling = scaling;
        return this;
      }

      Interface build() {
        checkState(!built, "This builder has already been used");
        checkNotNull(resource);
        checkNotNull(surfaceSpacing);
        checkNotNull(ruptureFloating);
        checkNotNull(ruptureScaling);
        this.built = true;
        return new Interface(this);
      }
    }
  }

  static class Grid extends SourceConfig {

    final double gridSpacing;
    final PointSourceType pointSourceType;
    final RuptureScaling ruptureScaling;
    final double maxDepth;
    final LogicTree<FocalMech> focalMechTree;
    final Map<String, GridDepthModel> gridDepthMap;

    private Grid(Builder builder) {
      super(builder.resource);
      this.gridSpacing = builder.gridSpacing;
      this.pointSourceType = builder.pointSourceType;
      this.ruptureScaling = builder.ruptureScaling;
      this.maxDepth = builder.maxDepth;
      this.focalMechTree = builder.focalMechTree;
      this.gridDepthMap = builder.gridDepthMap;
    }

    static Builder builder(Path resource) {
      return new Builder(resource);
    }

    static class Builder {

      private boolean built = false;

      private Path resource;
      private Double gridSpacing;
      private PointSourceType pointSourceType;
      private RuptureScaling ruptureScaling;
      private Double maxDepth;
      private LogicTree<FocalMech> focalMechTree;
      private Map<String, GridDepthModel> gridDepthMap;

      private Builder(Path resource) {
        this.resource = resource;
      }

      Builder spacing(double spacing) {
        this.gridSpacing = spacing;
        return this;
      }

      Builder sourceType(PointSourceType type) {
        this.pointSourceType = type;
        return this;
      }

      Builder scaling(RuptureScaling scaling) {
        this.ruptureScaling = scaling;
        return this;
      }

      Builder maxDepth(double depth) {
        this.maxDepth = depth;
        return this;
      }

      Builder focalMechTree(LogicTree<FocalMech> tree) {
        this.focalMechTree = tree;
        return this;
      }

      Builder gridDepthMap(Map<String, GridDepthModel> depthMap) {
        this.gridDepthMap = depthMap;
        return this;
      }

      Grid build() {
        checkState(!built, "This builder has already been used");
        checkNotNull(resource);
        checkNotNull(gridSpacing);
        checkNotNull(pointSourceType);
        checkNotNull(ruptureScaling);
        checkNotNull(maxDepth);
        checkNotNull(focalMechTree);
        checkNotNull(gridDepthMap);
        this.built = true;
        return new Grid(this);
      }
    }
  }
}
